
#ifndef GLOBAL_DATA
#define GLOBAL_DATA

#include "../../pbr.glsl"

#extension GL_EXT_nonuniform_qualifier : require

#define M_PI 3.14159265358979323846264338327950288

struct InputType {
    float px, py, pz;
    float nx, ny, nz;
    float uvx, uvy;
    float tx, ty, tz;

    uint material_index;
};

struct Vertex {
    vec3 position;
    vec3 normal;
    vec2 uv;
    vec3 tangent;
};

struct Material {
    vec4 color;

    int color_texture_index;
    int normal_texture_index;
    int emissive_texture_index;
    int roughness_texture_index;

    vec4 emissive_factor;

    float metallic_factor;
    float roughness_factor;
    uint alpha_mask;
    float alpha_cut_off;
};

struct Triangle {
    Vertex v0;
    Vertex v1;
    Vertex v2;

    Material material;

    vec3 barycentrics;
};

layout(set = 1, binding = 0) readonly buffer VertexBuffers {
    InputType vertices[ ];
} vertex_buffers;

layout (set = 2, binding = 0) readonly uniform MaterialBuffers {
    Material material;
} materials[ ];

layout (set = 3, binding = 0) uniform sampler2D textures[ ];

layout (set = 4, binding = 0) readonly buffer Light {
    vec3 position;
    float power;
    vec3 color;
    float arc;
    vec3 direction;
    float t_min;
} lights [ ];

layout (set = 5, binding = 0) uniform LightInfo {
    uint point_light_count;
    uint directional_light_count;
    uint spot_light_count;
} light_info;

// shaders helper functions
vec2 bary_lerp(vec2 a, vec2 b, vec2 c, vec3 barycentrics) {
    return a * barycentrics.x + b * barycentrics.y + c * barycentrics.z;
}

vec3 bary_lerp(vec3 a, vec3 b, vec3 c, vec3 barycentrics) {
    return a * barycentrics.x + b * barycentrics.y + c * barycentrics.z;
}

vec4 bary_lerp(vec4 a, vec4 b, vec4 c, vec3 barycentrics) {
    return a * barycentrics.x + b * barycentrics.y + c * barycentrics.z;
}

float intensity(float power, float distance) {
    if (distance == 0.0) {
        return 1.0;
    }

    // surface of a sphere
    // A = 4 * PI * r^2
    float surface = 4 * M_PI * distance * distance;

    return min(1.0, power / surface);
}

float rescale(float x, float x0, float x1, float y0, float y1) {
    return y0 + (y1 - y0) * (x - x0) / (x1 - x0);
}

vec3 rescale(vec3 v, float x0, float x1, float y0, float y1) {
    return vec3(
        rescale(v.x, x0, x1, y0, y1),
        rescale(v.y, x0, x1, y0, y1),
        rescale(v.z, x0, x1, y0, y1)
    );
}

Vertex get_vertex(const uint index, out uint material_index) {
    InputType t = vertex_buffers.vertices[index];

    Vertex v;

    v.position = vec3(t.px, t.py, t.pz);
    v.normal = vec3(t.nx, t.ny, t.nz);
    v.uv = vec2(t.uvx, t.uvy);
    v.tangent = vec3(t.tx, t.ty, t.tz);
    material_index = t.material_index;

    return v;
}

Triangle current_triangle(const uint primitive_id, vec2 attribs) {
    uint dummy;
    uint material_index;

    Triangle triangle;

    triangle.v0 = get_vertex(3 * primitive_id, material_index);
    triangle.v1 = get_vertex(3 * primitive_id + 1, dummy);
    triangle.v2 = get_vertex(3 * primitive_id + 2, dummy);

    triangle.material = materials[nonuniformEXT(material_index)].material;
    triangle.barycentrics = vec3(1.0f - attribs.x - attribs.y, attribs.x, attribs.y);

    return triangle;
}

vec2 uv(Triangle triangle) {
    return bary_lerp(triangle.v0.uv, triangle.v1.uv, triangle.v2.uv, triangle.barycentrics);
}

vec4 vertex_color(Vertex v, vec4 material_color) {
    if (v.uv.x == -1.0) {
        return vec4(v.uv.y, v.tangent);
    } else {
        return material_color;
    }
}

vec4 color(Triangle triangle) {
    if (triangle.v0.uv.x == -1.0) {
        vec4 c1 = vertex_color(triangle.v0, triangle.material.color);
        vec4 c2 = vertex_color(triangle.v1, triangle.material.color);
        vec4 c3 = vertex_color(triangle.v2, triangle.material.color);

        return bary_lerp(c1, c2, c3, triangle.barycentrics);
    } else if (triangle.material.color_texture_index == -1) {
        return triangle.material.color;
    } else {
        return texture(textures[nonuniformEXT(triangle.material.color_texture_index)], uv(triangle));
    }
}

vec3 normal(Triangle triangle) {
    vec3 normal = normalize(bary_lerp(triangle.v0.normal, triangle.v1.normal, triangle.v2.normal, triangle.barycentrics));

    if (triangle.material.normal_texture_index != -1) {
        vec3 tangent = normalize(bary_lerp(triangle.v0.tangent, triangle.v1.tangent, triangle.v2.tangent, triangle.barycentrics));
        vec3 B = normalize(cross(normal, tangent));
        mat3 TBN = mat3(tangent, B, normal);

        normal = TBN * normalize(texture(textures[nonuniformEXT(triangle.material.normal_texture_index)], uv(triangle)).xyz * 2.0 - vec3(1.0));
    }

    return normal;
}

void metallic_roughness(Triangle triangle, out float metallic_factor, out float roughness_factor) {
    if (triangle.material.roughness_texture_index == -1) {
        metallic_factor = triangle.material.metallic_factor;
        roughness_factor = triangle.material.roughness_factor;
    } else {
        vec4 tmp = texture(textures[nonuniformEXT(triangle.material.roughness_texture_index)], uv(triangle));

        // take channels as described by the glTF standard
        metallic_factor = tmp.b;
        roughness_factor = tmp.g;
    }
}

vec3 emissive_factor(Triangle triangle) {
    if (triangle.material.emissive_texture_index == -1) {
        return triangle.material.emissive_factor.xyz;
    } else {
        return texture(textures[nonuniformEXT(triangle.material.emissive_texture_index)], uv(triangle)).xyz;
    }
}

float random(inout vec2 st) {
    float v = fract(sin(dot(st, vec2(12.9898, 78.233))) * 43758.5453123);

    st.x = st.y;
    st.y = v;

    return v;
}

vec2 random_vec2(inout vec2 st) {
    return vec2(random(st), random(st));
}

vec3 random_vec3(inout vec2 st) {
    return vec3(random(st), random(st), random(st));
}

PBRInfo prepare(Triangle triangle, vec3 viewer_position, vec3 world_position) {
    vec4 color;

    if (triangle.material.color_texture_index == -1) {
        pbr_color(
            triangle.material.color,
            triangle.material.alpha_mask,
            triangle.material.alpha_cut_off,
            color
        );
    } else {
        pbr_color(
            triangle.material.color,
            textures[nonuniformEXT(triangle.material.color_texture_index)],
            uv(triangle),
            triangle.material.alpha_mask,
            triangle.material.alpha_cut_off,
            color
        );
    }

    float metallic_factor = triangle.material.metallic_factor;
    float roughness_factor = triangle.material.roughness_factor;

    if (triangle.material.roughness_texture_index == -1) {
        pbr_metallic_roughness(metallic_factor, roughness_factor);
    } else {
        pbr_metallic_roughness(
            metallic_factor,
            roughness_factor,
            textures[nonuniformEXT(triangle.material.roughness_texture_index)],
            uv(triangle)
        );
    }

    vec3 normal = normal(triangle);

    PBRInfo pbr_info = prepare(color, roughness_factor, metallic_factor, normal, viewer_position, world_position);

    return pbr_info;
}

#endif // GLOBAL_DATA
