use crate::scene::rendering::{
    ExtensionCheck, RenderingFrontEnd,
    rasterizer::{deferred::Rasterizer, traditional::TraditionalRasterizer},
    raytracer::{
        hardwareraytracer::HardwareRayTracer, raytracer::RayTracer,
        softwareraytracer::SoftwareRayTracer,
    },
    shared::shadercompiler::{ShaderCompiler, ShaderTypeConverter},
};
use crate::{include_shader, prelude::*};

#[cfg(feature = "timings")]
use super::super::timings::Timings;

use anyhow::Result;
use ecs::*;
use utilities::prelude::cgmath::{Matrix4, Vector3, vec3};

use std::sync::Mutex;
#[cfg(feature = "timings")]
use std::time::Instant;

use std::sync::Arc;
use std::time::Duration;

#[derive(Resource)]
pub struct Scene {
    screen_width: f32,
    screen_height: f32,

    render_type: SceneType,

    last_frame: Duration,
    frame_time: Duration,

    // renderer of this scene
    renderer: Box<dyn RenderingFrontEnd + Send + Sync>,

    frustum_check: Option<
        Box<
            dyn Fn(&EntityObject, &Matrix4<f32>, &ViewFrustum, &FrustumPlanes) -> Result<bool>
                + Send
                + Sync,
        >,
    >,

    particle_system_vulkan_objects: ParticleSystemVulkanObjects,

    #[cfg(feature = "timings")]
    timings: Timings,
}

impl Scene {
    pub(crate) fn new(
        rasterizer_info: RasterizerInfo,
        raytracer_info: RaytracingInfo,
        graphics_info: GraphicsInfo,
        world: &mut WorldBuilder,
    ) -> Result<()> {
        let context = world.resources.get::<Context>();

        let device = context.device();
        let queue = context.queue();
        let (screen_width, screen_height) = (context.width() as f32, context.height() as f32);
        let images = &context.images();

        let (renderer, render_type): (Box<dyn RenderingFrontEnd + Send + Sync>, SceneType) =
            Self::create_rendering_front_end(
                device,
                queue,
                (screen_width, screen_height),
                images,
                rasterizer_info,
                raytracer_info,
                graphics_info,
            )?;

        let compiler = ShaderCompiler::new([include_shader!(
            "../content/components/particle_system.comp"
        )])?;

        let cs = compiler.compile::<_, 0>(device, &[], |_| {})?;

        let particle_system_vulkan_objects = ParticleSystemVulkanObjects::new(device, queue, cs)?;

        let scene = Scene {
            screen_width,
            screen_height,

            render_type,

            frustum_check: None,

            renderer,

            last_frame: Duration::from_secs(0),
            frame_time: Duration::from_secs(0),

            particle_system_vulkan_objects,

            #[cfg(feature = "timings")]
            timings: Timings::default()
                .add_timing("fire events")
                .add_timing("move events")
                .add_timing("entity changes")
                .add_timing("particles")
                .add_timing("frustum")
                .add_timing("renderer"),
        };

        Self::insert_default_updates(world)?;
        world.resources.insert(scene);

        Ok(())
    }
}

impl Scene {
    pub(crate) fn _world_to_screen_space(
        (screen_width, screen_height): (f32, f32),
        world_space: Vector3<f32>,
        view: &View,
    ) -> Result<(i32, i32)> {
        let (view, proj) = {
            let proj = view.proj();
            let view = view.camera().old_view_matrix();

            (view, proj)
        };

        let p = proj * view * world_space.extend(1.0);

        let pos = vec3(p.x / p.w, p.y / p.w, p.z / p.w);

        let x = Self::screen_space_to_coord(screen_width, pos.x);
        let y = Self::screen_space_to_coord(screen_height, pos.y);

        Ok((x, y))
    }

    pub(crate) fn insert_default_updates(world: &mut WorldBuilder) -> Result<()> {
        world.add_update(
            "animate",
            5_500_000,
            |commands: &mut Commands, _entity, draw: &mut Draw, animation: &mut Animation| {
                animation.animate(commands.now(), draw)?;

                Ok(())
            },
            EmptyFilter,
        )?;

        world.add_update(
            "particle_system",
            5_000_000,
            |commands: &mut Commands,
             entity,
             particle_system: &mut ParticleSystem,
             scene: &mut Scene| {
                let now = commands.now();

                if !particle_system.update(now) {
                    commands.remove_entity(entity);
                    return Ok(());
                }

                {
                    let inheritance_info = CommandBuffer::inheritance_info(None, None, None, None);
                    let mut begin_info =
                        VkCommandBufferBeginInfo::new(VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT);
                    begin_info.set_inheritance_info(&inheritance_info);

                    let particle_handles = scene.particle_system_vulkan_objects_mut();
                    let pipeline = particle_handles.pipeline.clone();
                    let mut recorder = particle_handles.begin(begin_info)?;

                    particle_system.animate(&mut recorder, &pipeline, now)?;
                }

                Ok(())
            },
            EmptyFilter,
        )?;

        Ok(())
    }

    pub fn set_frustum_check<F>(&mut self, f: F)
    where
        F: Fn(&EntityObject, &Matrix4<f32>, &ViewFrustum, &FrustumPlanes) -> Result<bool>
            + Send
            + Sync
            + 'static,
    {
        self.frustum_check = Some(Box::new(f));
    }

    fn create_rendering_front_end(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        images: &TargetMode<Vec<Arc<Image>>>,
        rasterizer_info: RasterizerInfo,
        raytracer_info: RaytracingInfo,
        graphics_info: GraphicsInfo,
    ) -> Result<(Box<dyn RenderingFrontEnd + Send + Sync>, SceneType)> {
        let create =
            |render_type| -> Result<Option<(Box<dyn RenderingFrontEnd + Send + Sync>, SceneType)>> {
                match render_type {
                    SceneType::Rasterizer => {
                        if rasterizer_info.use_deferred {
                            match Rasterizer::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images,
                                graphics_info.render_scale,
                                rasterizer_info,
                                !Rasterizer::check(device),
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            };
                        } else {
                            match TraditionalRasterizer::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images,
                                rasterizer_info,
                                graphics_info,
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            };
                        }
                    }
                    SceneType::Raytracer => {
                        if RayTracer::<HardwareRayTracer>::check(device) {
                            match RayTracer::<HardwareRayTracer>::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images.mono().len(),
                                graphics_info,
                                raytracer_info,
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            }
                        }

                        if RayTracer::<SoftwareRayTracer>::check(device) {
                            match RayTracer::<SoftwareRayTracer>::new(
                                device,
                                queue,
                                (screen_width, screen_height),
                                images.mono().len(),
                                graphics_info,
                                raytracer_info,
                            ) {
                                Ok(r) => return Ok(Some((Box::new(r), render_type))),
                                Err(msg) => println!("{}", msg),
                            }
                        };
                    }
                }

                Ok(None)
            };

        match create(graphics_info.render_type)? {
            Some(scene) => Ok(scene),
            None => {
                if let Some(scene_type) = Self::check_scene_type(
                    device,
                    graphics_info.render_type,
                    rasterizer_info.use_deferred,
                ) {
                    return Ok(create(scene_type)?.unwrap());
                }

                Err(anyhow::anyhow!("failed to create renderer"))
            }
        }
    }

    pub fn check_scene_type(
        device: &Arc<Device>,
        preferred_scene_type: SceneType,
        use_deferred: bool,
    ) -> Option<SceneType> {
        let check_rasterizer = |device, use_deferred| {
            if use_deferred {
                if Rasterizer::check(device) {
                    return Some(SceneType::Rasterizer);
                }
            } else {
                return Some(SceneType::Rasterizer);
            }

            None
        };

        match preferred_scene_type {
            SceneType::Rasterizer => check_rasterizer(device, use_deferred),
            SceneType::Raytracer => {
                if RayTracer::<HardwareRayTracer>::check(device)
                    || RayTracer::<SoftwareRayTracer>::check(device)
                {
                    return Some(SceneType::Raytracer);
                }

                check_rasterizer(device, use_deferred)
            }
        }
    }

    #[inline]
    pub(crate) fn screen_space_to_coord(max: f32, v: f32) -> i32 {
        (((v + 1.0) / 2.0) * max) as i32
    }

    pub fn world_to_screen_space(&self, world_space: Vector3<f32>) -> Result<(i32, i32)> {
        Self::_world_to_screen_space(
            (self.screen_width, self.screen_height),
            world_space,
            self.view(),
        )
    }

    pub fn screen_space_to_world(&self, x: u32, y: u32) -> Result<Option<Vector3<f32>>> {
        let scale = self.renderer.render_scale();

        self.renderer
            .screen_to_world((x as f32 * scale) as u32, (y as f32 * scale) as u32)
    }

    pub fn particle_system_vulkan_objects_mut(&mut self) -> &mut ParticleSystemVulkanObjects {
        &mut self.particle_system_vulkan_objects
    }

    pub fn particle_system_vulkan_objects(&self) -> &ParticleSystemVulkanObjects {
        &self.particle_system_vulkan_objects
    }

    pub fn add_light(&mut self, light: Light) -> Result<()> {
        self.renderer.add_light(light)?;

        Ok(())
    }

    pub fn remove_light(&mut self, light: Light) -> Result<()> {
        self.renderer.remove_light(light)?;

        Ok(())
    }

    pub fn clear_lights(&mut self) -> Result<()> {
        self.renderer.clear_lights()?;

        Ok(())
    }

    pub fn view(&self) -> &View {
        self.renderer.view()
    }

    pub fn view_mut(&mut self) -> &mut View {
        self.renderer.view_mut()
    }

    pub fn render_type(&self) -> SceneType {
        self.render_type
    }

    pub fn frame_time(&self) -> Duration {
        self.frame_time
    }
}

impl TScene for Scene {
    fn process(
        &mut self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
        world: &mut World,
    ) -> anyhow::Result<()> {
        self.frame_time = {
            let now = world.now();

            let d = if self.last_frame < now {
                now - self.last_frame
            } else {
                Duration::from_secs(0)
            };

            self.last_frame = now;

            d
        };

        #[cfg(feature = "timings")]
        if let Some(timings) = self
            .timings
            .check_timing(self.now(), Duration::from_millis(2))
        {
            if !timings.is_empty() {
                println!("timings: \n{timings:#?}");
            }
        }

        // queue particle system commands
        {
            #[cfg(feature = "timings")]
            let before = Instant::now();

            self.particle_system_vulkan_objects
                .execute(buffer_recorder)?;

            #[cfg(feature = "timings")]
            self.timings
                .add("particles", Instant::now().duration_since(before));
        }

        // gather entities
        let content = {
            #[cfg(feature = "timings")]
            let before = Instant::now();

            let content = match &self.frustum_check {
                Some(frustum_check) => {
                    let view = self.renderer.view();
                    let vp = view.proj() * view.camera().old_view_matrix();

                    let frustum = view.frustum();
                    let planes = frustum.planes();

                    let content: Vec<Option<&EntityObject>> = world
                        .entities()
                        .map(|entity_object| {
                            if frustum_check(entity_object, &vp, &frustum, &planes)? {
                                Ok(Some(entity_object))
                            } else {
                                Ok(None)
                            }
                        })
                        .collect::<Result<Vec<Option<&EntityObject>>>>()?;

                    content.into_iter().flatten().collect()
                }
                None => world.entities().collect(),
            };

            #[cfg(feature = "timings")]
            self.timings
                .add("frustum", Instant::now().duration_since(before));

            content
        };

        // render
        {
            #[cfg(feature = "timings")]
            let before = Instant::now();

            self.renderer.process(
                world.had_entity_changes(),
                content,
                buffer_recorder,
                images,
                indices,
            )?;

            #[cfg(feature = "timings")]
            self.timings
                .add("renderer", Instant::now().duration_since(before));
        }

        Ok(())
    }

    fn resize(
        &mut self,
        window_width: f32,
        window_height: f32,
        images: &TargetMode<Vec<Arc<Image>>>,
        _world: &mut World,
    ) -> anyhow::Result<()> {
        self.screen_width = window_width;
        self.screen_height = window_height;

        self.renderer.resize(window_width, window_height, images)
    }
}
