use std::{fs, mem::ManuallyDrop};

use assetpath::AssetPath;
use mlua::{FromLua, FromLuaMulti, Function, IntoLua, IntoLuaMulti, Lua, Table as MLuaTable};

pub struct Globals<'a> {
    table: MLuaTable,
    lua: &'a Lua,
}

impl<'a> Globals<'a> {
    pub fn set<K: IntoLua, V: IntoLua>(&self, key: K, value: V) -> Result<(), anyhow::Error> {
        self.table.set(key, value)?;

        Ok(())
    }

    pub fn get<K: IntoLua, V: FromLua>(&self, key: K) -> Result<V, anyhow::Error> {
        Ok(self.table.get(key)?)
    }

    pub fn create_table(&self) -> Result<Table, anyhow::Error> {
        Ok(Table {
            table: self.lua.create_table()?,
        })
    }

    pub fn set_table(&self, name: &str, table: Table) -> Result<(), anyhow::Error> {
        self.set(name, table.table)?;

        Ok(())
    }

    pub fn get_table(&self, name: &str) -> Result<Table, anyhow::Error> {
        Ok(Table {
            table: self.get(name)?,
        })
    }
}

pub struct Table {
    table: MLuaTable,
}

impl Table {
    pub fn set<K: IntoLua, V: IntoLua>(&self, key: K, value: V) -> Result<(), anyhow::Error> {
        self.table.set(key, value)?;

        Ok(())
    }

    pub fn get<K: IntoLua, V: FromLua>(&self, key: K) -> Result<V, anyhow::Error> {
        Ok(self.table.get(key)?)
    }
}

pub struct LuaFunction {
    lua: Lua,
    function: ManuallyDrop<Function>,
}

impl LuaFunction {
    pub fn new(path: &AssetPath) -> Result<Self, anyhow::Error> {
        let lua = Lua::new();

        let function = ManuallyDrop::new(
            lua.load(fs::read_to_string(&path.full_path())?.as_bytes())
                .eval()?,
        );

        Ok(Self { lua, function })
    }

    pub fn globals<F, E>(&self, mut change_globals: F) -> Result<(), anyhow::Error>
    where
        F: FnMut(&Globals) -> Result<(), E>,
        E: std::error::Error + Send + Sync + 'static,
    {
        let globals = Globals {
            table: self.lua.globals(),
            lua: &self.lua,
        };

        change_globals(&globals)?;

        Ok(())
    }

    pub fn execute<A: IntoLuaMulti, R: FromLuaMulti>(&self, input: A) -> Result<R, anyhow::Error> {
        Ok(self.function.call(input)?)
    }
}

unsafe impl Sync for LuaFunction {}
unsafe impl Send for LuaFunction {}

impl Drop for LuaFunction {
    fn drop(&mut self) {
        // enforce function before dropping lua runtime
        // to manully ensure lifetime requirements
        unsafe {
            ManuallyDrop::drop(&mut self.function);
        }
    }
}
